@@ -4,7 +4,7 @@ JSONLD - A toolkit for interacting with JSON-LD data.
 
 =head1 VERSION
 
-This document describes JSONLD version 0.000_02
+This document describes JSONLD version 0.000_03
 
 =head1 SYNOPSIS
 
@@ -34,7 +34,7 @@ linked data.
 package JSONLD {
 	use v5.18;
 	use autodie;
-	our $VERSION	= '0.000_02';
+	our $VERSION	= '0.000_03';
 	use utf8;
 	use Moo;
 	use LWP;
@@ -50,7 +50,7 @@ package JSONLD {
 	use Data::Dumper;
 	use Clone 'clone';
 	use Carp qw(confess);
-	use B qw(svref_2object SVf_IOK SVf_POK);
+	use B qw(svref_2object SVf_IOK SVf_POK SVf_NOK SVf_IOK);
 	use namespace::clean;
 	
 	has 'base_iri' => (is => 'rw', required => 0, default => sub { IRI->new('http://example.org/') });
@@ -64,7 +64,7 @@ package JSONLD {
 	our $debug		= 0;
 	my %keywords	= map { $_ => 1 } qw(: @base @container @context @direction @graph @id @import @included @index @json @language @list @nest @none @prefix @propagate @protected @reverse @set @type @value @version @vocab);
 	
-=item C<< expand( $data ) >>
+=item C<< expand( $data, [expandContext => $ctx] ) >>
 
 Returns the JSON-LD expansion of C<< $data >>.
 
@@ -73,9 +73,17 @@ Returns the JSON-LD expansion of C<< $data >>.
 	sub expand {
 		my $self	= shift;
 		my $d		= shift;
+		my %args	= @_;
+		
 		my $ctx		= {
 			'@base' => $self->base_iri->abs, # TODO: not sure this follows the spec, but it's what makes test t0089 pass
 		};
+		if (my $ec = $args{expandContext}) {
+			if (ref($ec) eq 'HASH' and exists $ec->{'@context'}) {
+				$ec	= $ec->{'@context'};
+			}
+			$ctx	= $self->_4_1_2_ctx_processing($ctx, $ec);
+		}
 # 		warn "Expanding...";
 		return $self->_expand($ctx, undef, $d);
 	}
@@ -86,7 +94,26 @@ Returns the JSON-LD expansion of C<< $data >>.
 		return 0 if ref($v);
 		my $sv	= svref_2object(\$v);
 		my $flags	= $sv->FLAGS;
-		return $flags & SVf_POK;
+		my $is_string	= $flags & SVf_POK;
+		return $is_string;
+	}
+	
+	sub _is_integer {
+		my $v	= shift;
+		return 0 unless (_is_numeric($v));
+		my $i	= int($v);
+		my $f	= $v - $i;
+		return ($f == 0);
+	}
+
+	sub _is_numeric {
+		my $v	= shift;
+		return 0 unless defined($v);
+		return 0 if ref($v);
+		my $sv	= svref_2object(\$v);
+		my $flags	= $sv->FLAGS;
+		my $is_num	= (($flags & SVf_NOK) or ($flags & SVf_IOK));
+		return $is_num;
 	}
 	
 	sub _expand {
@@ -168,10 +195,18 @@ Returns the JSON-LD expansion of C<< $data >>.
 	sub _ctx_contains_protected_terms {
 		my $self	= shift;
 		my $ctx		= shift;
+		my @prot	= $self->_ctx_protected_terms($ctx);
+		return scalar(@prot);
+	}
+	
+	sub _ctx_protected_terms {
+		my $self	= shift;
+		my $ctx		= shift;
+		my @protected;
 		foreach my $term (keys %{ $ctx->{'terms'} }) {
-			return 1 if $ctx->{'terms'}{$term}{'protected'};
+			push(@protected, $term) if $ctx->{'terms'}{$term}{'protected'};
 		}
-		return 0;
+		return @protected;
 	}
 	
 	sub _is_node_object {
@@ -256,12 +291,15 @@ Returns the JSON-LD expansion of C<< $data >>.
 				# 5.1
 				println "5.1" if $debug;
 				if (not($override_protected) and $self->_ctx_contains_protected_terms($activeCtx)) {
-					println "5.1.1" if $debug;
+					my @prot	= $self->_ctx_protected_terms($activeCtx);
+					println "5.1.1 " . Data::Dumper->Dump([\@prot], ['protected_terms']) if $debug;
 					die 'invalid context nullification'; # 5.1.1
 				} else {
 					println "5.1.2 moving to next context" if $debug;
 					my $prev	= $result;
-					$result	= {};
+					$result	= {
+						'@base' => $self->base_iri->abs, # TODO: not sure this follows the spec, but it's what makes test t0089 pass
+					};
 					if ($propagate) {
 						$result->{'previous_context'}	= $prev;
 					}
@@ -554,7 +592,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 		if (not(defined($value))) {
 			println "7" if $debug;
 			$value	= {'@id' => undef}; # 7
-		} elsif (not(ref($value))) {
+		} elsif (_is_string($value)) {
 			# 8
 			println "8 value = {'\@id' => $value}" if $debug;
 			$value	= {'@id' => $value};
@@ -758,7 +796,6 @@ Returns the JSON-LD expansion of C<< $data >>.
 				} elsif (any { $_ eq '@set' } @$container and any { $_ =~ /^[@](index|graph|id|type|language)$/ } @$container) {
 					
 				} else {
-					warn Dumper($container) if $debug;
 					println "21.1(b)" if $debug;
 					die 'invalid container mapping';
 				}
@@ -951,10 +988,14 @@ Returns the JSON-LD expansion of C<< $data >>.
 		}
 		
 		my $property_scoped_ctx;
+		my $property_scoped_ctx_defined	= 0;
 		my $tdef = $self->_ctx_term_defn($activeCtx, $activeProp);
-		if ($tdef and my $lctx = $tdef->{'@context'}) {
-			println "3 property-scoped context for property $activeProp" if $debug;
+		warn Dumper($tdef);
+		if ($tdef and exists $tdef->{'@context'}) {
+			$property_scoped_ctx_defined	= 1;
+			my $lctx = $tdef->{'@context'};
 			$property_scoped_ctx	= $lctx; # 3
+			println "3 property-scoped context for property $activeProp: " . Data::Dumper->Dump([$property_scoped_ctx], [qw(context)]) if $debug;
 		}
 		
 		if (not(ref($element))) {
@@ -964,12 +1005,13 @@ Returns the JSON-LD expansion of C<< $data >>.
 				println "4.1 returning from _5_1_2_expansion: free floating scalar" if $debug;
 				return; # 4.1
 			}
-			if (defined($property_scoped_ctx)) {
+			if ($property_scoped_ctx_defined) {
 				println "4.2" if $debug;
 				$activeCtx = $self->_4_1_2_ctx_processing($activeCtx, $property_scoped_ctx); # 4.2
 				println "after 4.2: " . Data::Dumper->Dump([$activeCtx], ['activeCtx']) if $debug;
 			}
 			
+			println "4.3" if $debug;
 			my $v	= $self->_5_3_2_value_expand($activeCtx, $activeProp, $element);
 			local($Data::Dumper::Indent)	= 1;
 			println "4.3 returning from _5_1_2_expansion with " . Data::Dumper->Dump([$v], ['expandedValue']) if $debug;
@@ -1010,8 +1052,10 @@ Returns the JSON-LD expansion of C<< $data >>.
 			return \@result; # 5.3
 		}
 		
-		println "6" if $debug;
-		die "Unexpected non-map encountered during expansion: $element" unless (ref($element) eq 'HASH'); # 6; assert
+		unless (ref($element) eq 'HASH') {
+			println "6" if $debug;
+			die "Unexpected non-map encountered during expansion: $element";
+		}
 
 		if (my $prevCtx = $activeCtx->{'previous_context'}) {
 			unless ($fromMap) {
@@ -1025,7 +1069,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 			}
 		}
 		
-		if (defined($property_scoped_ctx)) {
+		if ($property_scoped_ctx_defined) {
 			println "8" if $debug;
 			my %args;
 			if ($tdef and exists $tdef->{'__source_base_iri'}) {
@@ -1123,7 +1167,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 				println "15.4" if $debug;
 				die 'invalid language-tagged value; ' . Dumper($result); # 15.4
 			} elsif (exists $result->{'@type'} and not($self->_is_iri($result->{'@type'}))) {
-				warn "Not an IRI \@type: " . Dumper($result->{'@type'});
+				println "Not an IRI \@type: " . Dumper($result->{'@type'});
 				println "15.5" if $debug;
 				die 'invalid typed value: ' . Dumper($result); # 15.5
 # 			} elsif (exists $result->{'@type'}) {
@@ -1164,11 +1208,13 @@ Returns the JSON-LD expansion of C<< $data >>.
 				println "18" if $debug;
 				return undef;
 			}
+			
+			println(Data::Dumper->Dump([$activeProp], ['activeProp'])) if $debug;
 			if (not(defined($activeProp)) or $activeProp eq '@graph') {
 				# 19
 				local($Data::Dumper::Indent)	= 0;
-				println "19 " . Data::Dumper->Dump([$result], ['*result']) if $debug;
-				if (ref($result) eq 'HASH' and scalar(@keys) == 0 or exists $result->{'@value'} or exists $result->{'@list'}) {
+				println "19 " . Data::Dumper->Dump([$result, \@keys], ['*result', '*keys']) if $debug;
+				if (ref($result) eq 'HASH' and (scalar(@keys) == 0 or exists $result->{'@value'} or exists $result->{'@list'})) {
 					println "19.1" if $debug;
 					$result	= undef; # 19.1
 				} elsif (ref($result) eq 'HASH' and scalar(@keys) == 1 and $keys[0] eq '@id') {
@@ -1270,7 +1316,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 
 				if ($expandedProperty eq '@type') {
 					println "13.4.4" if $debug;
-					my $is_string = not(ref($value));
+					my $is_string = _is_string($value);
 					my $is_array	= ref($value) eq 'ARRAY';
 					my $is_array_of_strings	= ($is_array and all { not(ref($_)) } @$value);
 					if (not($is_string) and not($is_array_of_strings)) {
@@ -1441,63 +1487,65 @@ Returns the JSON-LD expansion of C<< $data >>.
 
 					println "13.4.13.2" if $debug;
 					$expandedValue	= $self->_5_1_2_expansion($activeCtx, '@reverse', $value, frameExpansion => $frameExpansion, ordered => $ordered); # 13.4.13.1
-					println "13.4.13.2" . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
+					println "13.4.13.2 " . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
 					
-					if (ref($expandedValue) eq 'HASH' and exists $expandedValue->{'@reverse'}) { # NOTE: spec text does not assert that expandedValue is a map
-						println "13.4.13.3" if $debug;
-						foreach my $property (keys %{ $expandedValue->{'@reverse'} }) {
-							my $__indent	= indent();
-							println "13.4.13.3 [$property]" if $debug;
-							my $item	= $expandedValue->{'@reverse'}{$property};
-							if (not exists $result->{$property}) {
-								println "13.4.13.3.1" if $debug;
-								$result->{$property}	= [];
-							}
-							
-							println "13.4.13.3.2" if $debug;
-							push(@{ $result->{$property} }, $item);
-						}
-					}
-					
-					if (ref($expandedValue) eq 'HASH') { # NOTE: spec text does not assert that expandedValue is a map
-						my @keys	= grep { $_ ne '@reverse' } keys %$expandedValue;
-						if (scalar(@keys)) {
-							println "13.4.13.4" if $debug;
-						
-							if (not exists $result->{'@reverse'}) {
-								println "13.4.13.4.1" if $debug;
-								$result->{'@reverse'}	= {};
-							}
-						
-							println "13.4.13.4.2" if $debug;
-							my $reverse_map	= $result->{'@reverse'};
-						
-							println "13.4.13.4.3" if $debug;
-							foreach my $property (grep { $_ ne '@reverse' } keys %{ $expandedValue }) {
+					foreach my $ev (@{ $expandedValue }) { # https://github.com/w3c/json-ld-api/issues/294
+						my $expandedValue	= $ev;
+						if (ref($expandedValue) eq 'HASH' and exists $expandedValue->{'@reverse'}) { # NOTE: spec text does not assert that expandedValue is a map
+							println "13.4.13.3" if $debug;
+							foreach my $property (keys %{ $expandedValue->{'@reverse'} }) {
 								my $__indent	= indent();
-								println "13.4.13.4.3 [$property]" if $debug;
-								my $items	= $expandedValue->{$property};
+								println "13.4.13.3 [$property]" if $debug;
+								my $item	= $expandedValue->{'@reverse'}{$property};
+								if (not exists $result->{$property}) {
+									println "13.4.13.3.1" if $debug;
+									$result->{$property}	= [];
+								}
 							
-								println "13.4.13.4.3.1" if $debug;
-								foreach my $item (@$items) {
+								println "13.4.13.3.2" if $debug;
+								push(@{ $result->{$property} }, @$item);
+							}
+						}
+					
+						if (ref($expandedValue) eq 'HASH') { # NOTE: spec text does not assert that expandedValue is a map
+							my @keys	= grep { $_ ne '@reverse' } keys %$expandedValue;
+							if (scalar(@keys)) {
+								println "13.4.13.4" if $debug;
+						
+								if (not exists $result->{'@reverse'}) {
+									println "13.4.13.4.1" if $debug;
+									$result->{'@reverse'}	= {};
+								}
+						
+								println "13.4.13.4.2" if $debug;
+								my $reverse_map	= $result->{'@reverse'};
+						
+								println "13.4.13.4.3" if $debug;
+								foreach my $property (grep { $_ ne '@reverse' } keys %{ $expandedValue }) {
 									my $__indent	= indent();
-									if ($self->_is_value_object($item) or $self->_is_list_object($item)) {
-										println "13.4.13.4.3.1.1" if $debug;
-										die 'invalid reverse property value';
-									}
+									println "13.4.13.4.3 [$property]" if $debug;
+									my $items	= $expandedValue->{$property};
+							
+									println "13.4.13.4.3.1" if $debug;
+									foreach my $item (@$items) {
+										my $__indent	= indent();
+										if ($self->_is_value_object($item) or $self->_is_list_object($item)) {
+											println "13.4.13.4.3.1.1" if $debug;
+											die 'invalid reverse property value';
+										}
 								
-									if (not exists $reverse_map->{$property}) {
-										println "13.4.13.4.3.1.2" if $debug;
-										$reverse_map->{$property}	= [];
-									}
+										if (not exists $reverse_map->{$property}) {
+											println "13.4.13.4.3.1.2" if $debug;
+											$reverse_map->{$property}	= [];
+										}
 								
-									println "13.4.13.4.3.1.3" if $debug;
-									push(@{ $reverse_map->{$property} }, $item);
+										println "13.4.13.4.3.1.3" if $debug;
+										push(@{ $reverse_map->{$property} }, $item);
+									}
 								}
 							}
 						}
 					}
-					
 					println "13.4.13.5 going to next element key" if $debug;
 					next; # 13.4.13.5
 				}
@@ -1725,8 +1773,10 @@ Returns the JSON-LD expansion of C<< $data >>.
 			if ($self->_cm_contains($container_mapping, '@list') and not $self->_is_list_object($expandedValue)) {
 				# 13.11
 				println "13.11" if $debug;
-				my @values	= (ref($expandedValue) eq 'ARRAY') ? @$expandedValue : ($expandedValue);
-				$expandedValue	= { '@list' => \@values };
+				unless (ref($expandedValue) eq 'ARRAY') {
+					$expandedValue	= [$expandedValue];
+				}
+				$expandedValue	= { '@list' => $expandedValue };
 				println "13.11 resulting in " . Data::Dumper->Dump([$expandedValue], ['*expandedValue']) if $debug;
 			}
 
@@ -1799,8 +1849,6 @@ Returns the JSON-LD expansion of C<< $data >>.
 				if (ref($expandedValue) eq 'ARRAY') {
 					push(@{$result->{$expandedProperty}}, @$expandedValue); # 13.14.2
 				} elsif (ref($expandedValue)) {
-					# NOTE: I'm assuming that this is the intention of 13.14.2,
-					# but it isn't actually spelled out in the spec text.
 					println "setting result[$expandedProperty]" if $debug;
 					push(@{$result->{$expandedProperty}}, $expandedValue); # 13.14.2
 				}
@@ -1824,7 +1872,7 @@ Returns the JSON-LD expansion of C<< $data >>.
 		
 		# https://github.com/w3c/json-ld-api/issues/262
 		println "14" if $debug;
-		my @keys			= keys %$nests;
+		my @keys			= sort keys %$nests; # https://github.com/w3c/json-ld-api/issues/295
 		foreach my $nesting_key (@keys) {
 			delete $nests->{$nesting_key};
 			# 14
@@ -1950,9 +1998,11 @@ Returns the JSON-LD expansion of C<< $data >>.
 		} elsif ($documentRelative) {
 			# 8
 			println "8" if $debug;
-			my $base = $activeCtx->{'@base'} // $self->base_iri;
-			my $i = IRI->new(value => $value, base => $base);
-			$value	= $i->abs;
+			my $base = $activeCtx->{'@base'};
+			if (defined $base) {
+				my $i = IRI->new(value => $value, base => $base);
+				$value	= $i->abs;
+			}
 		}
 
 		println "9 returning from _5_2_2_iri_expansion with final value: $value" if $debug;
@@ -1960,23 +2010,24 @@ Returns the JSON-LD expansion of C<< $data >>.
 	}
 	
 	sub _5_3_2_value_expand {
-		println "ENTER    =================> _5_3_2_value_expand" if $debug;
-		my $__indent	= indent();
 		my $self		= shift;
 		my $activeCtx	= shift;
 		my $activeProp	= shift;
 		my $value		= shift;
+		my $_vs			= Data::Dumper->new([$value], ['value'])->Terse(1)->Dump([$value], ['value']);
+		println "ENTER    =================> _5_3_2_value_expand($_vs)" if $debug;
+		my $__indent	= indent();
 		
 		my $tdef	= $self->_ctx_term_defn($activeCtx, $activeProp);
 
 		if (exists $tdef->{'type_mapping'}) {
-			if ($tdef->{'type_mapping'} eq '@id' and not(ref($value))) {
+			if ($tdef->{'type_mapping'} eq '@id' and _is_string($value)) {
 				my $iri	= $self->_5_2_2_iri_expansion($activeCtx, $value, documentRelative => 1);
 				println "1 returning from _5_3_2_value_expand with new map containing \@id: $iri" if $debug;
 				return { '@id' => $iri }; # 1
 			}
 
-			if ($tdef->{'type_mapping'} eq '@vocab' and not(ref($value))) {
+			if ($tdef->{'type_mapping'} eq '@vocab' and _is_string($value)) {
 				my $iri	= $self->_5_2_2_iri_expansion($activeCtx, $value, vocab => 1, documentRelative => 1);
 				println "1 returning from _5_3_2_value_expand with new map containing vocab \@id: $iri" if $debug;
 				return { '@id' => $iri }; # 2
@@ -2057,11 +2108,18 @@ RDF-related methods:
 		$self->next_identifier_id(0);
 		$self->_7_2_2_nodemap_generation($expandedInput, $map);
 		println(Data::Dumper->Dump([$map], ['node_map'])) if $debug;
-
 		$self->_8_1_2_to_rdf($map, $dataset);
 		return $dataset;
 	}
 	
+	sub _is_well_formed_graph_node {
+		my $self	= shift;
+		my $value	= shift;
+		return 1 if $self->_is_well_formed_iri($value);
+		return 1 if ($value =~ /^_:/);
+		return 0;
+	}
+	
 	sub _is_well_formed_language {
 		my $self	= shift;
 		my $value	= shift;
@@ -2072,6 +2130,14 @@ RDF-related methods:
 		return $ok;
 	}
 	
+	sub _is_well_formed_datatype {
+		my $self	= shift;
+		my $value	= shift;
+		return 1 if ($value eq '@json');
+		return 1 if $self->_is_well_formed_iri($value);
+		return 0;
+	}
+	
 	sub _is_well_formed_iri {
 		my $self	= shift;
 		my $value	= shift;
@@ -2112,6 +2178,7 @@ RDF-related methods:
 		my $__indent	= indent();
 		my $self			= shift;
 		my $element			= shift;
+		println(Data::Dumper->Dump([$element], [qw(element)])) if $debug;
 		my $map				= shift;
 		my $activeGraph		= shift // '@default';
 		my $activeSubject	= shift;
@@ -2178,6 +2245,9 @@ RDF-related methods:
 						push(@{ $subjectNode->{$activeProp} }, $element);
 					}
 				}
+			} else {
+				println "4.2" if $debug;
+				push(@{ $list->{'@list'} }, $element);
 			}
 		} elsif (exists $element->{'@list'}) {
 			println "5" if $debug;
@@ -2185,10 +2255,11 @@ RDF-related methods:
 			my $result	= {'@list' => []};
 			
 			println "5.2" if $debug;
-			$self->_7_2_2_nodemap_generation($element->{'@list'}, $map, $activeGraph, $activeSubject, $activeProp, $list);
+			$self->_7_2_2_nodemap_generation($element->{'@list'}, $map, $activeGraph, $activeSubject, $activeProp, $result);
 			
 			if (not defined($list)) {
 				println "5.3" if $debug;
+				println(Data::Dumper->Dump([$result], [qw(result)])) if $debug;
 				push(@{ $subjectNode->{$activeProp} }, $result);
 			} else {
 				println "5.4" if $debug;
@@ -2247,12 +2318,14 @@ RDF-related methods:
 							push(@{ $subjectNode->{$activeProp} }, $reference);
 						}
 					}
+				} else {
+					push(@{ $list->{'@list'} }, $reference);
 				}
 			}
 			
 			if (exists $element->{'@type'}) {
 				println "6.7" if $debug;
-				my %exists	= map { $_ => 1 } @{ $element->{'@type'} };
+				my %exists	= map { $_ => 1 } @{ $node->{'@type'} };
 				push(@{ $node->{'@type'} }, grep { not exists $exists{$_} }@{ $element->{'@type'} });
 				delete $element->{'@type'};
 			}
@@ -2390,9 +2463,9 @@ RDF-related methods:
 						foreach my $type (@$values) {
 							if ($self->_is_well_formed_iri($type)) {
 								my $q	= $self->new_quad(
-									$subject,
+									$self->new_graph_node($subject),
 									$self->new_iri('http://www.w3.org/1999/02/22-rdf-syntax-ns#type'),
-									$type,
+									$self->new_iri($type),
 									$graph_iri
 								);
 								$self->add_quad($q, $dataset);
@@ -2426,7 +2499,8 @@ RDF-related methods:
 								$self->add_quad($q, $dataset);
 							}
 							println "1.3.2.5.4" if $debug;
-							foreach my $q (@$list_triples) {
+							foreach my $t (@$list_triples) {
+								my $q	= $t->as_quad($graph_iri);
 								$self->add_quad($q, $dataset);
 							}
 						}
@@ -2442,8 +2516,9 @@ RDF-related methods:
 		my $self			= shift;
 		my $item			= shift;
 		my $list_triples	= shift;
-		if ($self->_is_node_object($item) and not $self->_is_well_formed_iri($item->{'@id'})) {
+		if ($self->_is_node_object($item) and not $self->_is_well_formed_graph_node($item->{'@id'})) {
 			println "1" if $debug;
+			println "*** _8_2_2_object_to_rdf RETURNING NULL on " . Data::Dumper->Dump([$item], ['item']) if $debug;
 			return;
 		}
 
@@ -2468,13 +2543,15 @@ RDF-related methods:
 		println "5" if $debug;
 		my $datatype	= $item->{'@type'};
 
-		if (defined($datatype) and not $self->_is_well_formed_iri($datatype)) { # https://github.com/w3c/json-ld-api/issues/282
+		if (defined($datatype) and not $self->_is_well_formed_datatype($datatype)) { # https://github.com/w3c/json-ld-api/issues/282 ; https://github.com/w3c/json-ld-api/issues/298
 			println "6" if $debug;
+			println "*** _8_2_2_object_to_rdf RETURNING NULL" if $debug;
 			return;
 		}
 		
 		if (exists $item->{'@language'} and not $self->_is_well_formed_language($item->{'@language'})) {
 			println "7" if $debug;
+			println "*** _8_2_2_object_to_rdf RETURNING NULL" if $debug;
 			return;
 		}
 		
@@ -2485,8 +2562,21 @@ RDF-related methods:
 		}
 		
 		println "9 TODO bool? " . Data::Dumper->Dump([$value], ['value']) if $debug;
-		println "10/11 TODO numeric? " . Data::Dumper->Dump([$value], ['value']) if $debug;
 		
+		if (_is_numeric($value) and (not(_is_integer($value)) or (defined($datatype) and $datatype eq 'http://www.w3.org/2001/XMLSchema#double'))) {
+			println "10" if $debug;
+			$value	= sprintf('%E', $value);
+			$value	=~ s/0+E/0E/;
+			$value	=~ s/(\d)0E/$1E/;
+			$value	=~ s/E[+]0+/E0/;
+			$value	=~ s/E0+([1-9])/E$1/;
+			$datatype	= 'http://www.w3.org/2001/XMLSchema#double';
+		} elsif (_is_integer($value) or (_is_numeric($value) and (defined($datatype) and $datatype eq 'http://www.w3.org/2001/XMLSchema#integer'))) {
+			println "11" if $debug;
+			$value	= sprintf('%d', $value);
+			$datatype	= 'http://www.w3.org/2001/XMLSchema#integer';
+		}
+
 		if (not defined($datatype)) {
 			println "12" if $debug;
 			$datatype	= (exists $item->{'@language'}) ? 'http://www.w3.org/1999/02/22-rdf-syntax-ns#langString' : 'http://www.w3.org/2001/XMLSchema#string';
@@ -2541,10 +2631,65 @@ RDF-related methods:
 
 	sub _8_3_2_list_conversion {
 		println "ENTER    =================> _8_3_2_list_conversion" if $debug;
-		my $__indent	= indent();
-		die 'TODO 8.3.2 list_conversion';
+		my $__indent		= indent();
+		my $self			= shift;
+		my $list			= shift;
+		my $list_triples	= shift;
+		
+		my $NIL	= $self->new_iri('http://www.w3.org/1999/02/22-rdf-syntax-ns#nil');
+		if (scalar(@$list) == 0) {
+			println "1" if $debug;
+			return $NIL;
+		}
+		
+		println "2" if $debug;
+		my $bnodes	= [map { $self->new_blank() } @$list];
+		foreach my $i (0..$#{$bnodes}) {
+			my $subject	= $bnodes->[$i];
+			my $item	= $list->[$i];
+			println "3" if $debug;
+			println "3.1" if $debug;
+			my $embedded_triples	= [];
+			
+			println "3.2" if $debug;
+			my $object	= $self->_8_2_2_object_to_rdf($item, $embedded_triples);
+			
+			if (defined($object)) {
+				println "3.3" if $debug;
+				push(@$list_triples, $self->new_triple(
+					$subject,
+					$self->new_iri('http://www.w3.org/1999/02/22-rdf-syntax-ns#first'),
+					$object
+				));
+			}
+			
+			println "3.4" if $debug;
+			my $rest	= ($i == $#{$bnodes}) ? $NIL : $bnodes->[$i+1];
+			push(@$list_triples, $self->new_triple(
+				$subject,
+				$self->new_iri('http://www.w3.org/1999/02/22-rdf-syntax-ns#rest'),
+				$rest
+			));
+			
+			println "3.5" if $debug;
+			push(@$list_triples, @$embedded_triples);
+		}
+
+		println "4" if $debug;
+		my $head	= (scalar(@$bnodes)) ? $bnodes->[0] : $NIL;
+		return $head;
 	}
 	
+	sub new_graph_node {
+		my $self	= shift;
+		my $value	= shift;
+		if ($value =~ /_:(.*)/) {
+			return $self->new_blank($1);
+		} else {
+			return $self->new_iri($value);
+		}
+	}
+
 	sub new_graphname {
 		my $self	= shift;
 		Carp::cluck 'RDF term constructors must be overloaded in a JSONLD subclass';
